@@ -1,31 +1,14 @@
-#  Copyright (c) 2011, Karl Gyllstrom
-#  All rights reserved.
-#
-#  Redistribution and use in source and binary forms, with or without
-#  modification, are permitted provided that the following conditions are met:
-#
-#  1. Redistributions of source code must retain the above copyright notice,
-#     this list of conditions and the following disclaimer.
-#
-#  2. Redistributions in binary form must reproduce the above copyright notice,
-#     this list of conditions and the following disclaimer in the documentation
-#     and/or other materials provided with the distribution.
-#
-#     THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
-#     IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
-#     THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
-#     PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
-#     CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
-#     EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
-#     PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
-#     PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
-#     LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
-#     NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
-#     SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
-#
-#     The views and conclusions contained in the software and documentation are
-#     those of the authors and should not be interpreted as representing
-#     official policies, either expressed or implied, of the FreeBSD Project.
+'''
+
+    blargs command line parser
+    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
+
+    Easy argument parsing with powerful dependency support.
+    :copyright: (c) 2012 by Karl Gyllstrom
+    :license: BSD (see LICENSE.txt)
+
+'''
+
 
 from __future__ import print_function
 
@@ -50,40 +33,84 @@ else:
     from urlparse import urlparse
     import ConfigParser as cpars
 
-#iterkeys
-#def iterkeys(
+
+class Multidict(object):
+    def __init__(self, dictionary=None):
+        self._values = {}
+        if dictionary is not None:
+            # XXX should we make sure dictionary conforms to our schema?
+            self._values.update(dictionary)
+
+    def __delitem__(self, key):
+        return self._values.__delitem__(key)
+
+    def __contains__(self, key):
+        return self._values.__contains__(key)
+
+    def copy(self):
+        return Multidict(self._values.copy())
+
+    def get(self, key):
+        return self._values.get(key)
+
+    def __str__(self):
+        return self._values.__str__()
+
+    def overwrite(self, key, value):
+        self._values[key] = value
+
+    def __setitem__(self, key, value):
+        if key in self._values:
+            v = self._values[key]
+            if not isinstance(v, list):
+                v = [v]
+
+            value = v + [value]
+
+        self._values[key] = value
+
+    def __getitem__(self, key):
+        return self._values.__getitem__(key)
+
+    def __iter__(self):
+        return iteritems(self._values)
 
 
-def _ConfigCaster(f):
-    cfp = cpars.ConfigParser()
-    cfp.readfp(open(f))
-    for sec in cfp.sections():
-        for key, value in cfp.items(sec):
-            yield key, value
+class _ConfigCaster(object):
+    def __init__(self, parent):
+        self._parent = parent
+
+    def __call__(self, filename):
+        cfp = cpars.ConfigParser()
+        cfp.readfp(open(filename))
+        for sec in cfp.sections():
+            for key, value in cfp.items(sec):
+                yield key, value
 
 
-def _RangeCaster(value):
-    def raise_error():
-        raise FormatError(('%s is not range format: N:N+i, N-N+i, or N'
-                + ' N+i') % value)
+class _RangeCaster(object):
+    def __call__(self, value):
+        def raise_error():
+            raise FormatError(('%s is not range format: N:N+i, N-N+i, or N'
+                    + ' N+i') % value)
 
-    splitter = None
-    for char in (' :-'):
-        if char in value:
-            splitter = char
-            break
+        splitter = None
+        for char in (' :-'):
+            if char in value:
+                splitter = char
+                break
 
-    if splitter:
-        toks = value.split(splitter)
-    else:
-        toks = [value]
+        if splitter:
+            toks = value.split(splitter)
+        else:
+            toks = [value]
 
-    if not (1 <= len(toks) <= 3):
-        raise_error()
-    try:
-        return xrange(*[int(y) for y in toks])
-    except ValueError:
-        raise_error()
+        if not (1 <= len(toks) <= 3):
+            raise_error()
+        try:
+            return xrange(*[int(y) for y in toks])
+        except ValueError:
+            raise_error()
 
 
 class _DirectoryOpenerCaster(object):
@@ -103,14 +130,16 @@ class _DirectoryOpenerCaster(object):
         return name
 
 
-def _FileOpenerCaster(mode=None, buffering=None):
-    kw = {}
-    if mode is not None:
-        kw['mode'] = mode
-    if buffering is not None:
-        kw['buffering'] = buffering
+class _FileOpenerCaster(object):
+    def __init__(self, mode=None, buffering=None):
+        self._kw = {}
+        if mode is not None:
+            self._kw['mode'] = mode
+        if buffering is not None:
+            self._kw['buffering'] = buffering
 
-    return partial(open, **kw)
+    def __call__(self, *args):
+        return open(*args, **self._kw)
 
 
 # ---------- decorators ---------- #
@@ -216,6 +245,7 @@ class ConditionError(ArgumentError):
     ''' Condition not met. '''
 
     def __init__(self, argname, condition):
+        super(ConditionError, self).__init__()
         self.argname = argname
         self.condition = condition
 
@@ -247,7 +277,8 @@ class UnspecifiedArgumentError(ArgumentError):
     ''' User supplies argument that isn't specified. '''
 
     def __init__(self, arg):
-        super(UnspecifiedArgumentError, self).__init__('illegal option %s' % arg)
+        super(UnspecifiedArgumentError, self).__init__('illegal option %s' %
+                arg)
 
 
 class MultipleSpecifiedArgumentError(ArgumentError):
@@ -356,14 +387,25 @@ class CallableCondition(Condition):
         return c
 
     def _inner_satisfied(self, parsed):
-        newargs = []
-        for arg in (self._main, self._other):
-            if isinstance(arg, Option):
-                newargs.append(parsed.get(arg.argname))
+        operands = []
+        for item in (self._main, self._other):
+            if isinstance(item, Option):
+                v = parsed.get(item.argname)
+                if isinstance(v, list):
+                    v = [vi.getvalue() for vi in v]
+                else:
+                    v = [v.getvalue()]
             else:
-                newargs.append(arg)
+                v = [item]
 
-        return self._call(*newargs)
+            operands.append(v)
+
+        for x1 in operands[0]:
+            for x2 in operands[1]:
+                if not self._call(x1, x2):
+                    return False
+
+        return True
 
     def __repr__(self):
         x = self._main.argname
@@ -379,6 +421,18 @@ class CallableCondition(Condition):
         return x
 
 
+def _daisy_chained(f):
+    @wraps(f)
+    def newfunc(self, *args, **kw):
+        result = f(self, *args, **kw)
+        if result is not None:
+            return result
+
+        return self
+
+    return newfunc
+
+
 class Option(Condition):
     def __init__(self, argname, parser):
         ''' Do not construct directly, as it will not be tethered to a
@@ -390,6 +444,8 @@ class Option(Condition):
         self.argname = argname
         self._parser = parser
         self._conditions = []
+        self._allows_multiple = False
+        self._description = None
 
     def copy(self):
         c = super(Option, self).copy()
@@ -398,27 +454,30 @@ class Option(Condition):
         c._conditions = self._conditions
         return c
 
+    @_daisy_chained
+    def described_as(self, description):
+        self._description = description
+
+    @_daisy_chained
     def requires(self, *conditions):
         ''' Specifiy other options/conditions which this argument requires.
 
         :param conditions: required conditions
-        :type others: sequence of either :class:`Option` or :class:`Condition`s 
+        :type others: sequence of either :class:`Option` or :class:`Condition`s
         '''
+
         [self._parser._set_requires(self.argname, x) for x in conditions]
-        return self
-
-    def condition(self, func):
-        self._conditions.append(func)
-        return self
 
+    @_daisy_chained
     def conflicts(self, *conditions):
         ''' Specifiy other conditions which this argument conflicts with.
 
         :param conditions: conflicting options/conditions
-        :type conditions: sequence of either :class:`Option` or :class:`Condition`
+        :type conditions: sequence of either :class:`Option` or
+            :class:`Condition`
         '''
+
         [self._parser._set_conflicts(self.argname, x) for x in conditions]
-        return self
 
     def __str__(self):
         v = '%s%s' % (self._parser._double_prefix, self.argname)
@@ -427,6 +486,7 @@ class Option(Condition):
             v += '/%s%s' % (self._parser._single_prefix, alias)
         return v
 
+    @_daisy_chained
     def shorthand(self, alias):
         ''' Set shorthand for this argument. Shorthand arguments are 1
         character in length and are prefixed by a single '-'. For example:
@@ -440,13 +500,13 @@ class Option(Condition):
         '''
 
         self._parser._add_shorthand(self.argname, alias)
-        return self
 
+    @_daisy_chained
     def default(self, value):
         ''' Provide a default value for this argument. '''
         self._parser._set_default(self.argname, value)
-        return self
 
+    @_daisy_chained
     def environment(self):
         ''' Pull argument value from OS environment if unspecified. The case of
         the argument name, all lower, and all upper are all tried. For example,
@@ -455,7 +515,7 @@ class Option(Condition):
 
         >>> with Parser(locals()) as p:
         ...    p.int('port').environment()
-        
+
         Both command lines work:
 
         ::
@@ -473,52 +533,55 @@ class Option(Condition):
                 default = os.environ.get(self.argname.upper())
 
         if default is not None:
+            # XXX not tested
             return self.default(default)
 
-        return self
-
-    def name(self):
-        return self.argname
-
+    @_daisy_chained
     def cast(self, cast):
         ''' Provide a casting value for this argument. '''
-        self._parser._casts[self.argname] = cast
-        return self
 
+        self._parser._readers[self.argname] = Caster(
+                self._parser._readers[self.argname], cast)
+
+    @_daisy_chained
     def required(self):
         ''' Indicate that this argument is required. '''
 
         self._parser._set_required(self.argname)
-        return self
 
+    @_daisy_chained
     def if_(self, condition):
         ''' Argument is required if ``conditions``. '''
         self._parser._set_required(self.argname, [-condition])
-        return self
 
+    @_daisy_chained
     def unless(self, condition):
         ''' Argument is required unless ``conditions``. '''
 
         self._parser._set_required(self.argname, [condition])
-        return self
 
+    @_daisy_chained
     def unspecified_default(self):
         ''' Indicate that values passed without argument labels will be
         attributed to this argument. '''
 
         self._parser._set_unspecified_default(self.argname)
-        return self
 
+    @_daisy_chained
     def multiple(self):
         ''' Indicate that the argument can be specified multiple times. '''
-        self._parser._set_multiple(self.argname)
-        return self
+
+        self._allows_multiple = True
 
     # --- conditions
 
     def _inner_satisfied(self, parsed):
         v = parsed.get(self.argname)
-        return v is not None and v != False
+
+        if not isinstance(v, list):
+            v = [v]
+
+        return all(x.is_resolvable() for x in v)
 
     def _make_condition(self, func, other):
         return CallableCondition(func, self, other)
@@ -546,9 +609,6 @@ class Option(Condition):
 
     # -- private access methods
 
-    def _cast(self):
-        return self._parser._casts.get(self.argname)
-
     def _isrequired(self):
         return self in self._parser._required
 
@@ -556,7 +616,7 @@ class Option(Condition):
         return self._parser._conflicts.get(self, [])
 
     def _getreqs(self):
-        return self._parser._requires.get(self, [])
+       return self._parser._requires.get(self, [])
 
     def _alias(self):
         return self._parser._source_to_alias.get(self.argname)
@@ -582,69 +642,139 @@ class _ArgumentReader(object):
     def __init__(self, parent):
         self.value = _ArgumentReader.UNSPECIFIED
         self.parent = parent
+        self._default = _ArgumentReader.UNSPECIFIED
         self._init()
 
-    def consume_or_skip(self, arg):
-        raise NotImplementedError()
+    def fresh_copy(self):
+        return self.__class__(self.parent)
 
     def _init(self):
-        raise NotImplementedError()
+        pass
+
+    def activate(self):
+        pass
+
+    def _set_default(self, default):
+        self._default = default
+
+    def consume_or_skip(self, arg):
+        raise NotImplementedError
+
+    def is_specified(self):
+        return self.value is not _ArgumentReader.UNSPECIFIED
+
+    def is_resolvable(self):
+        return (self.is_specified() or self._default is not
+                _ArgumentReader.UNSPECIFIED)
+
+    def default(self):
+        if self._default is not _ArgumentReader.UNSPECIFIED:
+            return self._default
+        return self.__class__.class_default()
 
     @classmethod
-    def default(cls):
+    def class_default(cls):
         return _ArgumentReader.UNSPECIFIED
 
+    def _get(self):
+        raise NotImplementedError
+
+    def getvalue(self):
+        if self.is_specified():
+            return self._get()
+
+        return self.default()
+
 
 class _MultiWordArgumentReader(_ArgumentReader):
-    def _init(self):
-        self.value = []
-
     def consume_or_skip(self, arg):
         if self.parent._is_argument_label(arg):
             return False
 
+        if not self.is_specified():
+            self.value = []
+
         self.value.append(arg)
         return True
 
-    def get(self):
-        if len(self.value) == 0:
+    def _get(self):
+        if not self.is_specified() or len(self.value) == 0:
+            # XXX
             raise MissingValueError
+
         return ' '.join(self.value)
 
 
 class _FlagArgumentReader(_ArgumentReader):
     def _init(self):
+        self.value = False
+
+    def activate(self):
         self.value = True
 
+    def is_resolvable(self):
+        return self.value
+
     def consume_or_skip(self, arg):
         return False
 
-    def get(self):
+    def _get(self):
         return self.value
 
     @classmethod
-    def default(cls):
+    def class_default(cls):
         return False
 
 
 class _SingleWordReader(_ArgumentReader):
-    def _init(self):
-        pass
-
     def consume_or_skip(self, arg):
-        if self.value is not _ArgumentReader.UNSPECIFIED:
-
+        if self.is_specified():
             return False
 
         self.value = arg
         return True
 
-    def get(self):
-        if self.value is _ArgumentReader.UNSPECIFIED:
+    def _get(self):
+        if not self.is_specified():
             raise MissingValueError
         return self.value
 
 
+class Caster(object):
+    def __init__(self, reader, cast):
+        self._reader = reader
+        self._cast = cast
+
+    def fresh_copy(self):
+        # XXX does _cast need to be copied too?
+        return self.__class__(self._reader.fresh_copy(), self._cast)
+
+    def activate(self):
+        self._reader.activate()
+
+    def getvalue(self):
+        try:
+            v = self._reader.getvalue()
+            if v is _ArgumentReader.UNSPECIFIED:
+                return None
+
+            return self._cast(v)
+        except ValueError:
+            raise FormatError
+
+    def is_resolvable(self):
+        return self._reader.is_resolvable()
+
+    def consume_or_skip(self, arg):
+        return self._reader.consume_or_skip(arg)
+
+    def is_specified(self):
+        return self._reader.is_specified()
+
+    def _set_default(self, default):
+        self._reader._set_default(default)
+
+
 # ---------- Argument readers ---------- #
 
 
@@ -670,26 +800,16 @@ class Group(Option):
 
     def __str__(self):
         return ', '.join([str(name) for name in self._names])
-
-
-class FakeSystemExit(ValueError):
-    # Used for testing
-    pass
-
+            
 
 class Parser(object):
     ''' Command line parser. '''
 
-    def __init__(self, store=None, default_help=True):
+    def __init__(self, store=None):
         self._options = {}
         self._readers = {}
-        self._option_labels = {}
-        self._multiple = set()
-        self._casts = {}
-        self._defaults = {}
         self._extras = []
         self._unspecified_default = None
-        self.require_n = {}
 
         # dict of A (required) -> args that could replace A if A is not
         # specified
@@ -716,14 +836,13 @@ class Parser(object):
         # help message
         self._help_prefix = None
 
-        self._suppress_sys_exit = False
-        self._out = sys.stdout
+        self._sys_exit_error = SystemExit
+        self.out = sys.stdout # XXX not documented
 
         # set by user
         self._init_user_set(store)
 
-        if default_help:
-            self.flag('help').shorthand('h')
+        self.flag('help').shorthand('h').described_as('Print help message.')
 
     def _init_user_set(self, store=None):
         if store is None:
@@ -732,6 +851,9 @@ class Parser(object):
         self._namemaps = {}
         self._rnamemaps = {}
 
+    def _argument_exists(self, name_or_alias):
+        return name_or_alias in self._readers or name_or_alias in self._alias
+
     def set_help_prefix(self, message):
         ''' Indicate text to appear before argument list when the ``help``
         function is triggered. '''
@@ -751,7 +873,9 @@ class Parser(object):
         (e.g., -a). '''
 
         if self._double_prefix in flag:
-            raise ValueError('single_prefix cannot be superset of double_prefix')
+            raise ValueError(
+                    'single_prefix cannot be superset of double_prefix')
+
         self._single_prefix = flag
         return self
 
@@ -788,10 +912,10 @@ class Parser(object):
             ...   p.config('conf')
 
             Now, arg ``a`` can be specfied on the command line, or in the
-            configuration file passed to ``conf``.  For example:
+            configuration file passed to ``conf``. For example:
 
             ::
-            
+
                 python test.py --a 3
                 python test.py --conf myconfig.cfg
 
@@ -810,19 +934,21 @@ class Parser(object):
             would be ignored (as the developer did not create an ``x``
             argument).
 
-            If anything is specified on the command line, its value is not
-            taken from the config file. For example:
-
-            ::
-            
-                python test.py --a 3 --config myconfig.cfg
-
-            In this case, the value of ``a`` is 3 (from the command line) and
-            not 5 (from the config file).
+            If an argument is specified both on the command line and within the
+            config file, it must be indicated as allowing multiple (via
+            :py:meth:`.multiple`).
 
             '''
 
-        return self.str(name).cast(_ConfigCaster)
+        return self.str(name).cast(_ConfigCaster(self))
+
+    def enum(self, name, values):
+        arg = self.str(name)
+        cond = arg == values[0]
+        for v in values[1:]:
+            cond = cond.or_(arg == v)
+
+        return arg.requires(cond)
 
     def int(self, name):
         ''' Add integer argument. '''
@@ -864,10 +990,7 @@ class Parser(object):
               python test.py --values 0 10 3  # -> xrange(0, 10, 3)
         '''
 
-
-        result = self._add_option(name).cast(_RangeCaster)
-        self._set_reader(name, _MultiWordArgumentReader)
-        return result
+        return self.multiword(name).cast(_RangeCaster())
 
     def multiword(self, name):
         ''' Accepts multiple terms as an argument. For example:
@@ -884,7 +1007,7 @@ class Parser(object):
         '''
 
         result = self._add_option(name)
-        self._set_reader(name, _MultiWordArgumentReader)
+        self._set_reader(name, _MultiWordArgumentReader(self))
         return result
 
     def bool(self, name):
@@ -897,7 +1020,7 @@ class Parser(object):
         while an absence indicates false. No arguments. '''
 
         result = self._add_option(name)
-        self._set_reader(name, _FlagArgumentReader)
+        self._set_reader(name, _FlagArgumentReader(self))
         return result
 
     def file(self, name, mode=None, buffering=None):
@@ -911,7 +1034,6 @@ class Parser(object):
         ...     p.file('output_file', mode='w')
         ...
         ... output_file.write(input_file.read())
-        
         '''
 
         return self.multiword(name).cast(_FileOpenerCaster(mode, buffering))
@@ -933,7 +1055,7 @@ class Parser(object):
                 raise FormatError('%s not valid URL' % value)
             return value
 
-        self.str(name).cast(parse)
+        return self.str(name).cast(parse)
 
 # --- aggregate calls --- #
 
@@ -993,6 +1115,7 @@ class Parser(object):
                 item = self._options[item]
             newreplacements.append(item)
 
+        # XXX [] allows redundancy?
         self._required.setdefault(arg, []).extend(newreplacements)
 
     @localize
@@ -1009,16 +1132,13 @@ class Parser(object):
         self._alias[alias] = source
         self._source_to_alias[source] = alias
 
-    def _add_option(self, name, argument_label=None):
+    def _add_option(self, name):
         name = self._localize(name)
 
         if name in self._readers:
             raise ValueError('multiple types specified for %s' % name)
 
-        self._readers[name] = _SingleWordReader
-
-        if argument_label is not None:
-            self._option_labels[name] = argument_label
+        self._set_reader(name, _SingleWordReader(self))
 
         o = Option(name, self)
         self._options[name] = o
@@ -1064,14 +1184,14 @@ class Parser(object):
                 arg.startswith(self._double_prefix))
 
     def _parse(self, tokenized):
-        argument_value = None
-        parsed = {}
+        current_reader = None
+        parsed = Multidict()
 
         for arg in tokenized:
-            if argument_value is not None:
-                if argument_value.consume_or_skip(arg):
+            if current_reader is not None:
+                if current_reader.consume_or_skip(arg):
                     continue
-                argument_value = None
+                current_reader = None
 
             argument_name = None
 
@@ -1088,127 +1208,103 @@ class Parser(object):
                 argument_name = self._localize(arg)
 
                 if is_full:
-                    argument_name = self._options.get(argument_name).argname
-                    argument_value = self._readers.get(argument_name)
+                    argument_name = self._options.get(argument_name)
                 else:
-                    argument_name = self._options.get(self._alias.get(argument_name)).argname
-                    argument_value = self._readers.get(argument_name)
+                    argument_name = self._options.get(
+                            self._alias.get(argument_name))
 
-                if argument_value is None:
+                if argument_name is not None:
+                    argument_name = argument_name.argname
+
+                current_reader = self._readers.get(argument_name)
+
+                if current_reader is None:
                     raise UnspecifiedArgumentError(argument_name)
 
-                argument_value = argument_value(self)
+                current_reader = current_reader.fresh_copy()
+                current_reader.activate()
 
             elif self._unspecified_default is not None:
                 argument_name = self._unspecified_default
 
                 # push value onto _SingleWordReader
-                argument_value = _SingleWordReader(self)
-                argument_value.consume_or_skip(arg)
+                current_reader = _SingleWordReader(self)
+                current_reader.consume_or_skip(arg)
 
             if argument_name:
-                parsed.setdefault(argument_name,
-                        []).append(argument_value)
+                parsed[argument_name] = current_reader
             else:
                 self._extras.append(arg)
 
+        for k, v in parsed:
+            if not isinstance(v, list):
+                v = [v]
+            for item in v:
+                if not item.is_specified():
+                    raise MissingValueError
+
         return parsed
 
     def _help_if_necessary(self, processed):
         if 'help' in processed:
             self.print_help()
-            sys.exit(0)
+            raise self._sys_exit_error(0)
 
-    def _assign_defaults(self, parsed):
+    def _config_values(self, parsed):
+        pc = parsed.copy()
+        for key, value in parsed:
+            if isinstance(value, Caster) and isinstance(value._cast,
+                    _ConfigCaster):  # XXX ugly
+
+                for k, v in value.getvalue():
+                    current_reader = pc.get(k)
+
+                    if current_reader is None:
+                        # developer didn't specify this argument
+                        continue
+
+                    is_specified = current_reader.is_specified()
+                    if is_specified:
+                        current_reader = current_reader.fresh_copy()
+
+                    current_reader.consume_or_skip(v)
+
+                    if is_specified:
+                        pc[k] = current_reader
+
+                del pc[key]
+
+        return pc
+
+    def _assign(self, combined):
         assigned = {}
-        for key, value in iteritems(self._readers):
-            if key in parsed:
-                continue
-
-            cast = self._casts.get(key)
-            cast = cast if cast else lambda x: x
-
-            if key in self._defaults:
-                value = self._defaults[key]
-            else:
-                value = value.default()
-
-            assigned[key] = value
-
-        return assigned
-
-    def _assign_configs(self, parsed, valid_args):
-        argvalues = {}
-        for key, values in iteritems(parsed):
-            cast = self._casts.get(key)
-            if cast and cast == _ConfigCaster:
-                value = values[0].get()
-                for k, v in cast(value):
-                    if k in valid_args:
-                        argvalues[k] = v
-
-        return argvalues
-
-    def _assign(self, parsed):
-        argvalues = {}
-        argvalues.update(self._assign_defaults(parsed))
-        argvalues.update(self._assign_configs(parsed, set(argvalues.keys())))
-
-        for key, values in iteritems(parsed):
+        for key, values in combined:
             try:
-                if key not in self._multiple:
-                    value = values[0].get()
+                if not self._options[key]._allows_multiple:
+                    value = values.getvalue()
                 else:
-                    value = [v.get() for v in values]
+                    if not isinstance(values, list):
+                        values = [values]
+
+                    value = [v.getvalue() for v in values]
 
                 if value is _ArgumentReader.UNSPECIFIED:
-                    continue
+                    value = None
 
-                argvalues[key] = value
+                assigned[key] = value
 
             except MissingValueError:
                 raise MissingValueError('%s specified but missing given value'
                         % key)
 
-        assigned = {}
-        for key, value in iteritems(argvalues):
-            cast = self._casts.get(key)
-            if value is _ArgumentReader.UNSPECIFIED:
-                value = None
-            elif cast:
-                if cast == _ConfigCaster:
-                    continue
-
-                try:
-                    if isinstance(value, list):
-                        newv = []
-                        values = value
-                        for value in values:
-                            newv.append(cast(value))
-                        value = newv
-                    else:
-                        value = cast(value)
-                except FormatError:
-                    raise
-                except ValueError as e:
-                    raise FormatError('Cannot cast \'%s\' to %s' % (value, cast))
-
-            assigned[key] = value
-
         return assigned
 
     def _check_multiple(self, assigned):
-        for key, values in iteritems(assigned):
-            if len(values) > 1 and key not in self._multiple:
+        for key, values in assigned:
+            if isinstance(values, list) and not self._options[key]._allows_multiple:
                 raise MultipleSpecifiedArgumentError(('%s specified multiple' +
                     ' times') % self._options[key])
 
-    def _check_conditions(self, assigned):
-        for arg in iterkeys(assigned):
-            for cond in self._options[arg]._conditions:
-                if not cond(assigned):
-                    raise FailedConditionError()
-
     def _check_required(self, assigned):
         for arg, replacements in iteritems(self._required):
             missing = []
@@ -1245,7 +1341,6 @@ class Parser(object):
                         raise ConflictError(arg.argname, conflict.argname)
 
     def _verify(self, assigned):
-        self._check_conditions(assigned)
         self._check_required(assigned)
         self._check_dependencies(assigned)
         self._check_conflicts(assigned)
@@ -1268,15 +1363,25 @@ class Parser(object):
 
         return args
 
+    def _combine_with_defaults(self, user_args):
+        copy = Multidict(self._readers.copy())
+
+        for k, v in user_args:
+            copy.overwrite(k, v)
+
+        return copy
+
     def _process_command_line(self, args=None):
         try:
             args = self._get_args(args)
             tokenized = self._tokenize(args)
-            parsed = self._parse(tokenized)
-            self._help_if_necessary(parsed)
-            self._check_multiple(parsed)
-            assigned = self._assign(parsed)
-            self._verify(assigned)
+            user_args = self._parse(tokenized)
+            user_args = self._combine_with_defaults(user_args)
+            user_args = self._config_values(user_args)
+            self._check_multiple(user_args)
+            # XXX self._help_if_necessary(user_args)
+            self._verify(user_args)
+            assigned = self._assign(user_args)
             self._assign_to_store(assigned)
         except ArgumentError as e:
             raise e
@@ -1286,7 +1391,7 @@ class Parser(object):
         return self._store
 
     def emit(self, *args):
-        print(*args, file=self._out)
+        print(*args, file=self.out)
 
     def process_command_line(self, args=None):
         try:
@@ -1294,43 +1399,52 @@ class Parser(object):
         except ArgumentError as e:
             self.bail(e)
 
+    def _usage(self):
+        return ('Usage: %s ' % sys.argv[0]) + ' '.join('[%s]' %
+                self._label(value) for value in self._options.values())
+
+    def _print_table(self, t):
+        # XXX what about empty list?
+        column_max_lengths = [len(x) for x in t[0]]
+
+        for row in t[1:]:
+            column_max_lengths = [max(column_max_lengths[i], len(column)) for i, column
+                    in enumerate(row)]
+
+        for row in t:
+            line = ''
+            for i, column in enumerate(row):
+                fmt = '   %-' + str(column_max_lengths[i]) + 's'
+                line += fmt
+
+            self.emit(line % row)
+
     def bail(self, e):
         msg = []
         msg.append('Error: ' + str(e))
-        msg.append('usage: %s' % sys.argv[0])
-        msg.append(' '.join('[%s]' % self._label(value) for value in self._options.values()))
-        self.emit('\n'.join(msg))
-        if self._suppress_sys_exit:
-            raise FakeSystemExit
-        sys.exit(1)
+        msg.append(self._usage())
 
-    @localize
-    @_options_to_names
-    def _set_multiple(self, name):
-        self._multiple.add(name)
+        self.emit('\n'.join(msg))
+
+        raise self._sys_exit_error(1)
 
     @localize
     @_options_to_names
     def _set_default(self, name, value):
-        self._defaults[name] = value
+        self._readers[name]._set_default(value)
+    #    self._defaults[name] = value
 
     @_localize_all
     @_verify_args_exist
     @_names_to_options
     def _set_requires(self, a, b):
-        self._requires.setdefault(a, []).append(b)
-
-    @_localize_all
-    @_verify_args_exist
-    @_options_to_names
-    def set_requires_n_of(self, a, n, *others):
-        self.require_n[a] = (n, others)
+        self._requires.setdefault(a, set()).add(b)
 
     @_localize_all
     @_verify_args_exist
     @_names_to_options
     def _set_conflicts(self, a, b):
-        self._conflicts.setdefault(a, []).append(b)
+        self._conflicts.setdefault(a, set()).add(b)
 
     def __enter__(self):
         return self
@@ -1341,21 +1455,20 @@ class Parser(object):
             return True
         return False
 
-    def _option_label(self, name, opt):
-        label = self._option_labels.get(name)
-        if label is not None:
-            return label
+    def _option_label(self, reader):
+        if not isinstance(reader, Caster):
+            return 'option'
 
-        if opt._cast() is int:
+        if reader._cast is int:
             return 'int'
 
-        if opt._cast() is float:
+        if reader._cast is float:
             return 'float'
 
-        if opt._cast() is float:
+        if reader._cast is float:
             return 'float'
 
-        if opt._cast() is _RangeCaster:
+        if isinstance(reader._cast, _RangeCaster):
             return 'range'
 
         return 'option'
@@ -1364,38 +1477,45 @@ class Parser(object):
         pkey = str(opt)
 
         reader = self._readers[opt.argname]
-        if reader != _FlagArgumentReader:
-            pkey = '%s <%s>' % (pkey, self._option_label(opt.argname, opt))
+        if not isinstance(reader, _FlagArgumentReader):
+            pkey = '%s <%s>' % (pkey, self._option_label(reader))
 
         return pkey
 
     def print_help(self):
+        self.emit(self._usage())
+
         if self._help_prefix:
             self.emit(self._help_prefix)
 
-        self.emit('Arguments: (! denotes required argument)')
+        self.emit('Options: (! denotes required argument)')
         column_width = -1
-        for key, value in iteritems(self._options):
-            column_width = max(column_width, len(self._label(value)))
 
-        fmt = '   %-' + str(column_width) + 's'
+        labels = []
+        for key, opt in iteritems(self._options):
+            name = self._label(opt)
+            desc = ''
+            if opt._description is not None:
+                desc = opt._description
 
-        for key, value in iteritems(self._options):
-            msg = fmt % self._label(value)
+            if opt._isrequired():
+                name = '!' + name
 
-            if value._isrequired():
-                msg = '!' + msg[1:]
-
-            conflicts = value._getconflicts()
+            conflict_str = ''
+            conflicts = opt._getconflicts()
             if conflicts:
-                msg += ' (Conflicts with %s)' % ', '.join(str(item) for item in conflicts)
+                conflict_str = 'Conflicts with %s' % ', '.join(str(item) for item in
+                        conflicts)
 
-            reqs = value._getreqs()
+            requirement_str = ''
+            reqs = opt._getreqs()
             if reqs:
-                msg += ' (Requires %s)' % ', '.join(str(item) for item in reqs)
+                requirement_str = 'Requires %s' % ', '.join(str(item) for item in reqs)
 
-            self.emit(msg)
+            labels.append((name, desc, conflict_str, requirement_str))
+
+        self._print_table(labels)
 
 
 __all__ = ['Parser']
-__version__ = '0.2.22a'
+__version__ = '0.2.23b'
